from __future__ import annotations

import sys

import numpy as np
import serialtbx.detector.cspad
from serialtbx.detector import cspad
from serialtbx.detector.xtc import env_distance

from cctbx.eltbx import attenuation_coefficient
from libtbx.phil import parse
from scitbx.array_family import flex
from scitbx.matrix import col

from dxtbx.format.FormatXTC import FormatXTC, locator_str
from dxtbx.model import Detector, ParallaxCorrectedPxMmStrategy

cspad_locator_str = """
  cspad {
    detz_offset = None
      .type = float
      .help = Distance from back of detector rail to sample interaction region (CXI) \
              or actual detector distance (XPP/MFX)
    apply_gain_mask = True
      .type = bool
      .help = flag to indicate if gain should be applied to cspad data
    dark_correction = True
      .type = bool
      .help = flag to decide if dark correction should be done
    use_psana_calib = False
      .type = bool
      .help = Use the psana calibration
    common_mode = default
      .type = str
      .help = Common mode correction "default", "cspad_default", or "unbonded"\
              see https://confluence.slac.stanford.edu/display/PSDM/Common+mode+correction+algorithms\
              default means no common mode corrections... the other two are psana corrections
    }
"""

cspad_locator_scope = parse(cspad_locator_str + locator_str, process_includes=True)


class FormatXTCCspad(FormatXTC):
    def __init__(self, image_file, locator_scope=cspad_locator_scope, **kwargs):
        super().__init__(image_file, locator_scope=locator_scope, **kwargs)
        assert (
            self.params.cspad.detz_offset is not None
        ), "Supply a detz_offset for the cspad"
        self._cache_psana_pedestals()  # NOTE: move to base FormatXTC class
        self._psana_gain_map_cache = {}

    @staticmethod
    def understand(image_file):
        try:
            params = FormatXTC.params_from_phil(cspad_locator_scope, image_file)
        except Exception:
            return False
        return any("cspad" in src.lower() for src in params.detector_address)

    def _get_psana_gain_map(self, run):
        """
        checks if user wants gain applied and caches a gain map per run
        """
        if run.run() not in self._psana_gain_map_cache:
            if self.params.cspad.apply_gain_mask:
                self._psana_gain_map_cache[run.run()] = (
                    self._get_psana_detector(run).gain_mask(run) > 0
                )
            else:
                self._psana_gain_map_cache[run.run()] = None

    def _cache_psana_pedestals(self):
        """Store a pedestal for each psana detector instance"""
        self._pedestals = {}
        for run_number, run in self._psana_runs.items():
            det = self._get_psana_detector(run)
            self._pedestals[run_number] = det.pedestals(run)

    def get_raw_data(self, index=None):
        if index is None:
            index = 0
        assert len(self.params.detector_address) == 1
        d = self.get_detector(index)
        event = self._get_event(index)
        run_number = event.run()
        run = self._psana_runs[run_number]
        det = self._get_psana_detector(run)
        data = cspad.get_psana_corrected_data(
            det,
            event,
            use_default=self.params.cspad.use_psana_calib,
            dark=self._pedestals[run_number]
            if self.params.cspad.dark_correction
            else False,
            common_mode=self.params.cspad.common_mode,
            apply_gain_mask=self.params.cspad.apply_gain_mask,
            gain_mask_value=None,
            per_pixel_gain=False,
            gain_mask=self._get_psana_gain_map(run),
        )
        data = data.astype(np.float64)
        self._raw_data = []
        for quad_count, quad in enumerate(d.hierarchy()):
            for sensor_count, sensor in enumerate(quad):
                for asic_count, asic in enumerate(sensor):
                    fdim, sdim = asic.get_image_size()
                    asic_data = data[
                        sensor_count + quad_count * 8,
                        :,
                        asic_count * fdim : (asic_count + 1) * fdim,
                    ]  # 8 sensors per quad
                    self._raw_data.append(flex.double(np.array(asic_data)))
        assert len(d) == len(self._raw_data)
        return tuple(self._raw_data)

    def get_detector(self, index=None):
        return FormatXTCCspad._detector(self, index)

    # XXX Implement recursive version
    def _detector(self, index=None):
        if index is None:
            index = 0

        run = self.get_run_from_index(index)
        det = self._get_psana_detector(run)
        geom = det.pyda.geoaccess(run.run())
        cob = cspad.read_slac_metrology(geometry=geom, include_asic_offset=True)
        distance = env_distance(
            self.params.detector_address[0], run.env(), self.params.cspad.detz_offset
        )
        d = Detector()
        pg0 = d.hierarchy()
        # first deal with D0
        det_num = 0
        origin = col((cob[(0,)] * col((0, 0, 0, 1)))[0:3])
        fast = col((cob[(0,)] * col((1, 0, 0, 1)))[0:3]) - origin
        slow = col((cob[(0,)] * col((0, 1, 0, 1)))[0:3]) - origin
        origin += col((0.0, 0.0, -distance))
        pg0.set_local_frame(fast.elems, slow.elems, origin.elems)
        pg0.set_name("D%d" % (det_num))
        for quad_num in range(4):
            # Now deal with Qx
            pg1 = pg0.add_group()
            origin = col((cob[(0, quad_num)] * col((0, 0, 0, 1)))[0:3])
            fast = col((cob[(0, quad_num)] * col((1, 0, 0, 1)))[0:3]) - origin
            slow = col((cob[(0, quad_num)] * col((0, 1, 0, 1)))[0:3]) - origin
            pg1.set_local_frame(fast.elems, slow.elems, origin.elems)
            pg1.set_name("D%dQ%d" % (det_num, quad_num))
            for sensor_num in range(8):
                # Now deal with Sy
                pg2 = pg1.add_group()
                origin = col((cob[(0, quad_num, sensor_num)] * col((0, 0, 0, 1)))[0:3])
                fast = (
                    col((cob[(0, quad_num, sensor_num)] * col((1, 0, 0, 1)))[0:3])
                    - origin
                )
                slow = (
                    col((cob[(0, quad_num, sensor_num)] * col((0, 1, 0, 1)))[0:3])
                    - origin
                )
                pg2.set_local_frame(fast.elems, slow.elems, origin.elems)
                pg2.set_name("D%dQ%dS%d" % (det_num, quad_num, sensor_num))
                # Now deal with Az
                for asic_num in range(2):
                    val = "ARRAY_D0Q%dS%dA%d" % (quad_num, sensor_num, asic_num)
                    p = pg2.add_panel()
                    origin = col(
                        (cob[(0, quad_num, sensor_num, asic_num)] * col((0, 0, 0, 1)))[
                            0:3
                        ]
                    )
                    fast = (
                        col(
                            (
                                cob[(0, quad_num, sensor_num, asic_num)]
                                * col((1, 0, 0, 1))
                            )[0:3]
                        )
                        - origin
                    )
                    slow = (
                        col(
                            (
                                cob[(0, quad_num, sensor_num, asic_num)]
                                * col((0, 1, 0, 1))
                            )[0:3]
                        )
                        - origin
                    )
                    p.set_local_frame(fast.elems, slow.elems, origin.elems)
                    p.set_pixel_size((cspad.pixel_size, cspad.pixel_size))
                    p.set_image_size(cspad.asic_dimension)
                    p.set_trusted_range(
                        (
                            serialtbx.detector.cspad.cspad_min_trusted_value,
                            serialtbx.detector.cspad.cspad_saturated_value,
                        )
                    )
                    p.set_name(val)

        try:
            beam = self._beam(index)
        except Exception:
            print(
                "No beam object initialized. Returning CSPAD detector without parallax corrections"
            )
            return d

        # take into consideration here the thickness of the sensor also the
        # wavelength of the radiation (which we have in the same file...)
        wavelength = beam.get_wavelength()
        thickness = 0.5  # mm, see Hart et al. 2012

        table = attenuation_coefficient.get_table("Si")
        # mu_at_angstrom returns cm^-1
        mu = table.mu_at_angstrom(wavelength) / 10.0  # mu: mm^-1
        t0 = thickness
        for panel in d:
            panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
        return d


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatXTCCspad.understand(arg))



from __future__ import annotations

import random

import pytest

from scitbx import matrix

from dxtbx.model import Panel


def tst_intersection_at_origin(intersection, wavelength, origin):
    eps = 1e-7

    # Set the beam vector to directed towards the detector origin
    s = (1.0 / wavelength) * origin

    # Get the x, y pixel coordinates
    x, y = intersection(s)

    # Check that they are equal to the detector origin
    assert abs(x) < eps
    assert abs(y) < eps


def tst_intersection_at_corners(intersection, panel, wavelength):
    eps = 1e-7

    # Get parameters from detector
    origin = matrix.col(panel.get_origin())
    fast_axis = matrix.col(panel.get_fast_axis())
    slow_axis = matrix.col(panel.get_slow_axis())
    image_size = panel.get_image_size_mm()

    # Set the beam vector to directed towards the detector origin
    s1 = (1.0 / wavelength) * origin
    s2 = (1.0 / wavelength) * (origin + fast_axis * image_size[0])
    s3 = (1.0 / wavelength) * (origin + slow_axis * image_size[1])
    s4 = (1.0 / wavelength) * (
        origin + fast_axis * image_size[0] + slow_axis * image_size[1]
    )

    # Get the x, y pixel coordinates
    x1, y1 = intersection(s1)
    x2, y2 = intersection(s2)
    x3, y3 = intersection(s3)
    x4, y4 = intersection(s4)

    # Check that they are equal to the detector origin
    assert abs(x1 - 0) < eps and abs(y1 - 0) < eps
    assert abs(x2 - image_size[0]) < eps and abs(y2 - 0) < eps
    assert abs(x3 - 0) < eps and abs(y3 - image_size[1]) < eps
    assert abs(x4 - image_size[0]) < eps and abs(y4 - image_size[1]) < eps


def tst_intersection_away_from_panel(intersection, panel, wavelength):
    # Set the beam vector to directed away from the detector origin
    s = -(1.0 / wavelength) * matrix.col(panel.get_origin())

    # Check that they are equal to None
    with pytest.raises(RuntimeError):
        intersection(s)


def test_beam_plane_intersection():
    # The input parameters (from a GXPARM.XDS file)
    fast_axis = (0.000000, -0.939693, -0.342020)
    slow_axis = (1.000000, 0.000000, 0.000000)
    normal = (0.000000, -0.342020, 0.939693)
    pixel_size = (0.172000, 0.172000)
    pixel_origin = (244.836136, 320.338531)
    image_size = (487, 619)
    distance = 122.124901
    wavelength = 0.689400

    # Calculate the vector to the detector (0, 0) pixel
    origin = (
        (matrix.col(fast_axis).normalize() * pixel_size[0]) * (0 - pixel_origin[0])
        + (matrix.col(slow_axis).normalize() * pixel_size[1]) * (0 - pixel_origin[1])
        + (distance * matrix.col(normal).normalize())
    )

    # Create a detector object
    panel = Panel(
        "", "", fast_axis, slow_axis, origin, pixel_size, image_size, (0, 0), 0.0, ""
    )

    # Create the intersection object
    intersection = panel.get_ray_intersection

    # Perform a load of tests
    tst_intersection_at_origin(intersection, wavelength, origin)
    tst_intersection_at_corners(intersection, panel, wavelength)
    tst_intersection_away_from_panel(intersection, panel, wavelength)


def tst_transform_at_origin(transform, panel):
    eps = 1e-7

    # Get lab coordinate at point
    x, y, z = transform((0, 0))

    # Check coordinate is at origin
    assert abs(x - panel.get_origin()[0]) < eps
    assert abs(y - panel.get_origin()[1]) < eps
    assert abs(z - panel.get_origin()[2]) < eps


def tst_transform_at_corners(transform, panel):
    eps = 1e-7

    # Get parameters from detector
    origin = matrix.col(panel.get_origin())
    fast_axis = matrix.col(panel.get_fast_axis())
    slow_axis = matrix.col(panel.get_slow_axis())
    image_size = panel.get_image_size_mm()

    # Set the beam vector to directed towards the detector origin
    xyz1 = origin
    xyz2 = origin + fast_axis * image_size[0]
    xyz3 = origin + slow_axis * image_size[1]
    xyz4 = origin + fast_axis * image_size[0] + slow_axis * image_size[1]

    # Get the x, y pixel coordinates
    xyz11 = matrix.col(transform((0, 0)))
    xyz22 = matrix.col(transform((image_size[0], 0)))
    xyz33 = matrix.col(transform((0, image_size[1])))
    xyz44 = matrix.col(transform((image_size[0], image_size[1])))

    # Check that they are equal to the detector origin
    assert abs(xyz11 - xyz1) < eps
    assert abs(xyz22 - xyz2) < eps
    assert abs(xyz33 - xyz3) < eps
    assert abs(xyz44 - xyz4) < eps


def test_plane_to_lab_transform():
    # The input parameters (from a GXPARM.XDS file)
    fast_axis = (0.000000, -0.939693, -0.342020)
    slow_axis = (1.000000, 0.000000, 0.000000)
    normal = (0.000000, -0.342020, 0.939693)
    pixel_size = (0.172000, 0.172000)
    pixel_origin = (244.836136, 320.338531)
    image_size = (487, 619)
    distance = 122.124901

    # Calculate the vector to the detector (0, 0) pixel
    origin = (
        (matrix.col(fast_axis).normalize() * pixel_size[0]) * (0 - pixel_origin[0])
        + (matrix.col(slow_axis).normalize() * pixel_size[1]) * (0 - pixel_origin[1])
        + (distance * matrix.col(normal).normalize())
    )

    # Create a detector object
    panel = Panel(
        "", "", fast_axis, slow_axis, origin, pixel_size, image_size, (0, 0), 0.0, ""
    )

    # Create the intersection object
    transform = panel.get_lab_coord

    # Perform a load of tests
    tst_transform_at_origin(transform, panel)
    tst_transform_at_corners(transform, panel)


def tst_fwd_rev_random(intersection, transform, panel):
    eps = 1e-7

    # Set the shift parameters
    image_size = panel.get_image_size_mm()

    def random_xy():
        return (random.random() * image_size[0], random.random() * image_size[1])

    # Loop a number of times
    num = 1000
    for i in range(num):

        # Create a detector coordinate
        xy = matrix.col(random_xy())

        # Calculate the lab coordinate of the vector
        s = transform(xy)

        # Calculate the detector coordinate
        xy_2 = intersection(s)

        # Check the vectors are almost equal
        assert abs(xy - matrix.col(xy_2)) < eps


def test_forward_and_reverse_transform():
    # The input parameters (from a GXPARM.XDS file)
    fast_axis = (0.000000, -0.939693, -0.342020)
    slow_axis = (1.000000, 0.000000, 0.000000)
    normal = (0.000000, -0.342020, 0.939693)
    pixel_size = (0.172000, 0.172000)
    pixel_origin = (244.836136, 320.338531)
    image_size = (487, 619)
    distance = 122.124901

    # Calculate the vector to the detector (0, 0) pixel
    origin = (
        (matrix.col(fast_axis).normalize() * pixel_size[0]) * (0 - pixel_origin[0])
        + (matrix.col(slow_axis).normalize() * pixel_size[1]) * (0 - pixel_origin[1])
        + (distance * matrix.col(normal).normalize())
    )

    # Create a detector object
    panel = Panel(
        "", "", fast_axis, slow_axis, origin, pixel_size, image_size, (0, 0), 0.0, ""
    )

    # Create the intersection object and transform object
    intersection = panel.get_ray_intersection
    transform = panel.get_lab_coord

    # Do a test
    tst_fwd_rev_random(intersection, transform, panel)



from __future__ import annotations

import os
from copy import deepcopy

import pycbf

import libtbx.phil
from cctbx.eltbx import attenuation_coefficient
from scitbx import matrix

from dxtbx.model.detector_helpers import (
    detector_helper_sensors,
    find_gain_value,
    find_undefined_value,
    find_underload_value,
    set_detector_distance,
    set_fast_slow_beam_centre_mm,
    set_mosflm_beam_centre,
)

try:
    from ..dxtbx_model_ext import (
        Detector,
        Panel,
        ParallaxCorrectedPxMmStrategy,
        SimplePxMmStrategy,
    )

except ModuleNotFoundError:
    from dxtbx_model_ext import (  # type: ignore
        Detector,
        Panel,
        ParallaxCorrectedPxMmStrategy,
        SimplePxMmStrategy,
    )

# N.B. this should probably be generalized for non
# flat detectors, or composite detectors constructed from a number of flat
# elements.


dxtbx_overload_scale = float(os.getenv("DXTBX_OVERLOAD_SCALE", "1"))

detector_phil_scope = libtbx.phil.parse(
    """
  detector
    .expert_level = 1
    .short_caption = "Detector overrides"
  {
    panel
      .multiple = True
    {
      id = 0
        .type = int
        .help = "The panel number"
        .short_caption = "Panel ID"

      name = None
        .type = str
        .help = "Override the panel name"
        .short_caption = "Panel name"

      type = None
        .type = str
        .help = "Override the panel type"
        .short_caption = "Panel type"

      gain = None
        .type = float(value_min=0)
        .help = "The gain of the detector panel"
        .short_caption = "Gain value"

      pedestal = None
        .type = float
        .help = "The pedestal of the detector panel"
        .short_caption = "Pedestal value"

      pixel_size = None
        .type = floats(size=2)
        .help = "Override the panel pixel size"
        .short_caption = "Panel pixel size"

      image_size = None
        .type = ints(size=2)
        .help = "Override the panel image size"
        .short_caption= "Panel image size"

      trusted_range = None
        .type = floats(size=2)
        .help = "Override the panel trusted range: [min-trusted-value, max-trusted-value]"
        .short_caption = "Panel trusted range"

      thickness = None
        .type = float
        .help = "Override the panel thickness"
        .short_caption = "Panel thickness"

      material = None
        .type = str
        .help = "Override the panel material"
        .short_caption = "Panel material"

      fast_axis = None
        .type = floats(size=3)
        .help = "Override the panel fast axis. Requires slow_axis and origin."
        .short_caption = "Panel fast axis direction"

      slow_axis = None
        .type = floats(size=3)
        .help = "Override the panel slow axis. Requires fast_axis and origin."
        .short_caption = "Panel slow axis direction"

      origin = None
        .type = floats(size=3)
        .help = "Override the panel origin. Requires fast_axis and slow_axis."
        .short_caption = "Panel origin vector"

      parallax_correction = None
        .type = bool
        .help = "Enable parallax correction. By default in overwrite mode, the"
                "value of None does nothing."
        .short_caption = "Enable parallax correction"
    }

    hierarchy
      .expert_level = 2
      {

      name = None
        .type = str
        .help = "Override the group name"
        .short_caption = "group name"

      fast_axis = None
        .type = floats(size=3)
        .help = "Override the panel fast axis. Requires slow_axis and origin."
        .short_caption = "Panel fast axis direction"

      slow_axis = None
        .type = floats(size=3)
        .help = "Override the panel slow axis. Requires fast_axis and origin."
        .short_caption = "Panel slow axis direction"

      origin = None
        .type = floats(size=3)
        .help = "Override the panel origin. Requires fast_axis and slow_axis."
        .short_caption = "Panel origin vector"

      group
        .multiple = True
      {
        id = None
          .type = ints
          .help = "The group identifier specifying the place in the hierarchy"
          .short_caption = "Group ID"

        name = None
          .type = str
          .help = "Override the group name"
          .short_caption = "Group name"

        fast_axis = None
        .type = floats(size=3)
          .help = "Override the group fast axis. Requires slow_axis and origin."
          .short_caption = "Group fast axis direction"

        slow_axis = None
          .type = floats(size=3)
          .help = "Override the group slow axis. Requires fast_axis and origin."
          .short_caption = "Group slow axis direction"

        origin = None
          .type = floats(size=3)
          .help = "Override the group origin. Requires fast_axis and slow_axis."
          .short_caption = "Group origin vector"

        panel = None
          .multiple = True
          .type = int
          .help = "The panel id"
          .short_caption = "Panel ID"
      }
    }

    mosflm_beam_centre = None
      .type = floats(size=2)
      .help = "Override the beam centre from the image headers, following "
              "the mosflm convention."
      .short_caption = "Beam centre coordinates (mm, mm) using the Mosflm convention"

    distance = None
      .type = float
      .help = "The detector distance (used when mosflm_beam_centre is set)"
      .short_caption = "Detector distance"

    fast_slow_beam_centre = None
      .type = floats(size_min=2, size_max=3)
      .help = "Override the beam centre from the image headers."
              "The first two values are the fast and slow pixel coordinate."
              "If the third is supplied it specifies a panel number."
      .short_caption = "Beam centre coordinates (px fast, px slow, [panel id])"

    slow_fast_beam_centre = None
      .type = ints(size_min=2, size_max=3)
      .help = "Alternative to fast_slow_beam_centre in which the coordinates"
              "are given in order (px slow, px fast, [panel id]). If"
              "fast_slow_beam_centre is set it will take priority over any"
              "values set here."
      .short_caption = "Beam centre coordinates (px slow, px fast, [panel id])"
  }
"""
)


def merge_panel_scope_extracts_by_id(panel_params):

    id_to_params = {}
    for (i, params) in enumerate(panel_params):
        if params.id not in id_to_params:
            id_to_params[params.id] = [
                i,
            ]
        else:
            id_to_params[params.id].append(i)

    merged_params = []
    for params_set in id_to_params.values():
        params0 = deepcopy(panel_params[params_set[0]])
        for i in params_set[1:]:
            params1 = deepcopy(panel_params[i])
            for key in params1.__dict__:
                if (
                    key.startswith("_")
                    or key == "id"
                    or params0.__dict__[key] == params1.__dict__[key]
                ):
                    continue
                if (
                    params0.__dict__[key] is not None
                    and params1.__dict__[key] is not None
                ):
                    raise RuntimeError(
                        f"Multiple definitions for {key} for panel id={params0.id}"
                    )
                if params0.__dict__[key] is None:
                    params0.__dict__[key] = params1.__dict__[key]
        merged_params.append(params0)
    return merged_params


class DetectorFactory:
    """A factory class for detector objects, which will encapsulate standard
    detector designs to make it a little easier to get started with these. In
    cases where a CBF image is provided a full description can be used, in
    other cases assumptions will be made about the experiment configuration.
    In all cases information is provided in the CBF coordinate frame."""

    @staticmethod
    def generate_from_phil(params, beam=None):
        """
        Create a new detector model from phil parameters

        """
        detector = Detector()

        # Create a list of panels
        panel_list = {}

        # merge panel params by id first
        merged = merge_panel_scope_extracts_by_id(params.detector.panel)

        for panel_params in merged:
            panel = Panel()
            if panel_params.name is not None:
                panel.set_name(panel_params.name)
            if panel_params.type is not None:
                panel.set_type(panel_params.type)
            if panel_params.gain is not None:
                panel.set_gain(panel_params.gain)
            if panel_params.pedestal is not None:
                panel.set_pedestal(panel_params.pedestal)
            if panel_params.pixel_size is not None:
                panel.set_pixel_size(panel_params.pixel_size)
            else:
                raise RuntimeError("No pixel size set")
            if panel_params.image_size is not None:
                panel.set_image_size(panel_params.image_size)
            else:
                raise RuntimeError("No image size set")
            if panel_params.trusted_range is not None:
                panel.set_trusted_range(panel_params.trusted_range)
            else:
                raise RuntimeError("No trusted range set")
            if panel_params.thickness is not None:
                panel.set_thickness(panel_params.thickness)
            if panel_params.material is not None:
                panel.set_material(panel_params.material)
            if panel_params.parallax_correction is True:
                if panel_params.material is None:
                    raise RuntimeError("No material for parallax correction")
                if panel_params.thickness is None:
                    raise RuntimeError("No thickness for parallax correction")
                if beam is None:
                    raise RuntimeError("No beam for parallax correction")
                table = attenuation_coefficient.get_table(panel_params.material)
                mu = table.mu_at_angstrom(beam.get_wavelength()) / 10.0
                t0 = panel_params.thickness
                panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
            if panel_params.fast_axis is None:
                panel_params.fast_axis = (1, 0, 0)
            if panel_params.slow_axis is None:
                panel_params.slow_axis = (0, 1, 0)
            if panel_params.origin is None:
                panel_params.origin = (0, 0, 0)
            panel.set_local_frame(
                panel_params.fast_axis, panel_params.slow_axis, panel_params.origin
            )
            if panel_params.id in panel_list:
                raise RuntimeError("Multiple panels with id=%d" % panel_params.id)
            panel_list[panel_params.id] = panel

        # Create the hierarchy
        panel_counter = 0
        root = detector.hierarchy()
        if params.detector.hierarchy.name is not None:
            root.set_name(params.detector.hierarchy.name)
        if params.detector.hierarchy.fast_axis is None:
            params.detector.hierarchy.fast_axis = (1, 0, 0)
        if params.detector.hierarchy.slow_axis is None:
            params.detector.hierarchy.slow_axis = (0, 1, 0)
        if params.detector.hierarchy.origin is None:
            params.detector.hierarchy.origin = (0, 0, 0)
        root.set_frame(
            params.detector.hierarchy.fast_axis,
            params.detector.hierarchy.slow_axis,
            params.detector.hierarchy.origin,
        )

        def get_parent(node, index):
            if len(index) == 0:
                return node
            return get_parent(node[index[0]], index[1:])

        for group_params in params.detector.hierarchy.group:
            parent = get_parent(root, group_params.id[:-1])
            assert len(parent) == group_params.id[-1]
            group = parent.add_group()
            if group_params.name is not None:
                group.set_name(group_params.name)
            if group_params.fast_axis is None:
                group_params.fast_axis = (1, 0, 0)
            if group_params.slow_axis is None:
                group_params.slow_axis = (0, 1, 0)
            if group_params.origin is None:
                group_params.origin = (0, 0, 0)
            group.set_local_frame(
                group_params.fast_axis, group_params.slow_axis, group_params.origin
            )
            for panel_id in group_params.panel:
                assert panel_id == panel_counter
                group.add_panel(panel_list[panel_id])
                panel_counter += 1
        if panel_counter == 0:
            for panel_id in range(max(panel_list.keys()) + 1):
                detector.add_panel(panel_list[panel_id])
        elif panel_counter != len(panel_list):
            raise RuntimeError("Inconsistent number of panels in hierarchy")

        # Return detector
        return detector

    @staticmethod
    def overwrite_from_phil(params, detector, beam=None):
        """
        Overwrite from phil parameters

        """
        # merge panel params by id first
        merged = merge_panel_scope_extracts_by_id(params.detector.panel)

        # Override any panel parameters
        for panel_params in merged:
            panel = detector[panel_params.id]
            if panel_params.name is not None:
                panel.set_name(panel_params.name)
            if panel_params.type is not None:
                panel.set_type(panel_params.type)
            if panel_params.gain is not None:
                panel.set_gain(panel_params.gain)
            if panel_params.pedestal is not None:
                panel.set_pedestal(panel_params.pedestal)
            if panel_params.pixel_size is not None:
                panel.set_pixel_size(panel_params.pixel_size)
            if panel_params.image_size is not None:
                panel.set_image_size(panel_params.image_size)
            if panel_params.trusted_range is not None:
                panel.set_trusted_range(panel_params.trusted_range)
            if panel_params.thickness is not None:
                panel.set_thickness(panel_params.thickness)
            if panel_params.material is not None:
                panel.set_material(panel_params.material)
            if panel_params.parallax_correction is None:
                if isinstance(
                    panel.get_px_mm_strategy(), ParallaxCorrectedPxMmStrategy
                ):
                    panel_params.parallax_correction = True
                else:
                    panel_params.parallax_correction = False
            if panel_params.parallax_correction is True:
                if beam is None:
                    raise RuntimeError("No beam for parallax correction")
                table = attenuation_coefficient.get_table(panel.get_material())
                mu = table.mu_at_angstrom(beam.get_wavelength()) / 10.0
                t0 = panel.get_thickness()
                panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
            else:
                panel.set_px_mm_strategy(SimplePxMmStrategy())
            axes = [panel_params.fast_axis, panel_params.slow_axis, panel_params.origin]
            if axes.count(None) != 3:
                if panel_params.fast_axis is None:
                    panel_params.fast_axis = panel.get_local_fast_axis()
                if panel_params.slow_axis is None:
                    panel_params.slow_axis = panel.get_local_slow_axis()
                if panel_params.origin is None:
                    panel_params.origin = panel.get_local_origin()
                panel.set_local_frame(
                    panel_params.fast_axis, panel_params.slow_axis, panel_params.origin
                )

        # Create the hierarchy
        if params.detector.hierarchy is not None:
            root = detector.hierarchy()
            if params.detector.hierarchy.name is not None:
                root.set_name(params.detector.hierarchy.name)
            if (
                params.detector.hierarchy.fast_axis is not None
                or params.detector.hierarchy.slow_axis is not None
                or params.detector.hierarchy.origin is not None
            ):
                if params.detector.hierarchy.fast_axis is None:
                    params.detector.hierarchy.fast_axis = root.get_fast_axis()
                if params.detector.hierarchy.slow_axis is None:
                    params.detector.hierarchy.slow_axis = root.get_slow_axis()
                if params.detector.hierarchy.origin is None:
                    params.detector.hierarchy.origin = root.get_origin()
                root.set_frame(
                    params.detector.hierarchy.fast_axis,
                    params.detector.hierarchy.slow_axis,
                    params.detector.hierarchy.origin,
                )

            def get_group(node, index):
                if len(index) == 0:
                    return node
                return get_group(node[index[0]], index[1:])

            for group_params in params.detector.hierarchy.group:
                group = get_group(root, group_params.id)
                if group_params.name is not None:
                    group.set_name(group_params.name)
                if (
                    group_params.fast_axis is not None
                    or group_params.slow_axis is not None
                    or group_params.origin is not None
                ):
                    if group_params.fast_axis is None:
                        group_params.fast_axis = group.get_local_fast_axis()
                    if group_params.slow_axis is None:
                        group_params.slow_axis = group.get_local_slow_axis()
                    if group_params.origin is None:
                        group_params.origin = group.get_local_origin()
                    group.set_local_frame(
                        group_params.fast_axis,
                        group_params.slow_axis,
                        group_params.origin,
                    )
                if len(group_params.panel) != 0:
                    raise RuntimeError("Can't reassign panels in groups")

        return detector

    @staticmethod
    def from_phil(params, reference=None, beam=None):
        """
        Convert phil parameters into detector model

        """
        # Check the input. If no reference detector is provided then
        # Create the detector model from scratch from the parameters
        if reference is None:
            detector = DetectorFactory.generate_from_phil(params, beam)
        else:
            detector = DetectorFactory.overwrite_from_phil(params, reference, beam)

        # If the distance is set
        if params.detector.distance is not None:
            set_detector_distance(detector, params.detector.distance)

        # If the mosflm beam centre is set then update
        if params.detector.mosflm_beam_centre is not None:
            assert beam is not None
            set_mosflm_beam_centre(detector, beam, params.detector.mosflm_beam_centre)

        fast_slow_beam_centre = None
        if (
            params.detector.fast_slow_beam_centre
            and params.detector.slow_fast_beam_centre
        ):
            raise RuntimeError(
                "Can only set one of fast_slow_beam_centre or slow_fast_beam_centre"
            )

        if params.detector.slow_fast_beam_centre:
            fast_slow_beam_centre = list(params.detector.slow_fast_beam_centre)
            fast_slow_beam_centre[0], fast_slow_beam_centre[1] = (
                fast_slow_beam_centre[1],
                fast_slow_beam_centre[0],
            )

        if params.detector.fast_slow_beam_centre:
            fast_slow_beam_centre = params.detector.fast_slow_beam_centre

        # If the fast slow beam centre is set then update
        if fast_slow_beam_centre is not None:
            panel_id = 0
            if len(fast_slow_beam_centre) > 2:
                if fast_slow_beam_centre[2].is_integer():
                    panel_id = int(fast_slow_beam_centre[2])
                else:
                    raise TypeError(
                        f"Panel ID received float: {fast_slow_beam_centre[2]}"
                    )
            if panel_id >= len(detector):
                raise IndexError(f"Detector does not have panel index {panel_id}")
            px_size_f, px_size_s = detector[0].get_pixel_size()
            fast_slow_beam_centre_mm = (
                fast_slow_beam_centre[0] * px_size_f,
                fast_slow_beam_centre[1] * px_size_s,
            )
            assert beam is not None
            set_fast_slow_beam_centre_mm(
                detector, beam, fast_slow_beam_centre_mm, panel_id=panel_id
            )

        return detector

    @staticmethod
    def from_dict(d, t=None):
        """Convert the dictionary to a detector model

        Params:
            d The dictionary of parameters
            t The template dictionary to use

        Returns:
            The detector model
        """
        if d is None and t is None:
            return None
        joint = t.copy() if t else {}
        if isinstance(d, list):
            d = {"panels": d}
        joint.update(d)

        # Create the model from the joint dictionary
        return Detector.from_dict(joint)

    @staticmethod
    def make_detector(
        stype,
        fast_axis,
        slow_axis,
        origin,
        pixel_size,
        image_size,
        trusted_range=(0.0, 1.0e6),
        px_mm=None,
        name="Panel",
        thickness=0.0,
        material="",
        mu=0.0,
        gain=None,
        pedestal=None,
        identifier="",
    ):
        """Ensure all types are correct before creating c++ detector class."""

        if px_mm is None:
            px_mm = SimplePxMmStrategy()
        d = Detector()
        p = d.add_panel()
        p.set_type(str(stype))
        p.set_name(str(name))
        p.set_local_frame(
            tuple(map(float, fast_axis)),
            tuple(map(float, slow_axis)),
            tuple(map(float, origin)),
        )
        p.set_pixel_size(tuple(map(float, pixel_size)))
        p.set_image_size(tuple(map(int, image_size)))
        p.set_trusted_range(tuple(map(float, trusted_range)))
        p.set_thickness(thickness)
        p.set_material(material)
        p.set_mu(mu)
        p.set_px_mm_strategy(px_mm)
        p.set_identifier(identifier)
        if gain is not None:
            p.set_gain(gain)
        if pedestal is not None:
            p.set_pedestal(pedestal)
        return d

    @staticmethod
    def simple(
        sensor,
        distance,
        beam_centre,
        fast_direction,
        slow_direction,
        pixel_size,
        image_size,
        trusted_range=(0.0, 0.0),
        mask=[],
        px_mm=None,
        mu=0.0,
        gain=None,
        pedestal=None,
        identifier="",
    ):
        """Construct a simple detector at a given distance from the sample
        along the direct beam presumed to be aligned with -z, offset by the
        beam centre - the directions of which are given by the fast and slow
        directions, which are themselves given as +x, +y, -x, -y. The pixel
        size is given in mm in the fast and slow directions and the image size
        is given in pixels in the same order. Everything else is the same as
        for the main reference frame."""

        assert fast_direction in ["-x", "+y", "+x", "-y"]
        assert slow_direction in ["-x", "+y", "+x", "-y"]

        assert fast_direction[1] != slow_direction[1]

        direction_map = {
            "+x": (1.0, 0.0, 0.0),
            "-x": (-1.0, 0.0, 0.0),
            "+y": (0.0, 1.0, 0.0),
            "-y": (0.0, -1.0, 0.0),
        }

        fast = matrix.col(direction_map[fast_direction])
        slow = matrix.col(direction_map[slow_direction])

        origin = (
            matrix.col((0, 0, -1)) * distance
            - fast * beam_centre[0]
            - slow * beam_centre[1]
        )

        detector = DetectorFactory.make_detector(
            DetectorFactory.sensor(sensor),
            fast,
            slow,
            origin,
            pixel_size,
            image_size,
            trusted_range,
            px_mm,
            mu=mu,
            gain=gain,
            pedestal=pedestal,
            identifier=identifier,
        )
        detector[0].mask = mask
        return detector

    @staticmethod
    def two_theta(
        sensor,
        distance,
        beam_centre,
        fast_direction,
        slow_direction,
        two_theta_direction,
        two_theta_angle,
        pixel_size,
        image_size,
        trusted_range=(0.0, 0.0),
        mask=[],
        px_mm=None,
        mu=0.0,
        gain=None,
        pedestal=None,
        identifier="",
    ):
        """Construct a simple detector at a given distance from the sample
        along the direct beam presumed to be aligned with -z, offset by the
        beam centre - the directions of which are given by the fast and slow
        directions, which are themselves given as +x, +y, -x, -y. The pixel
        size is given in mm in the fast and slow directions and the image size
        is given in pixels in the same order. Everything else is the same as
        for the main reference frame. Also given are the direction of the
        two-theta axis and the angle in degrees by which the detector is
        moved."""

        assert fast_direction in ["-x", "+y", "+x", "-y"]
        assert slow_direction in ["-x", "+y", "+x", "-y"]
        assert two_theta_direction in ["-x", "+y", "+x", "-y"]

        assert fast_direction[1] != slow_direction[1]

        direction_map = {
            "+x": (1.0, 0.0, 0.0),
            "-x": (-1.0, 0.0, 0.0),
            "+y": (0.0, 1.0, 0.0),
            "-y": (0.0, -1.0, 0.0),
        }

        fast = matrix.col(direction_map[fast_direction])
        slow = matrix.col(direction_map[slow_direction])

        origin = (
            matrix.col((0, 0, -1)) * distance
            - fast * beam_centre[0]
            - slow * beam_centre[1]
        )

        two_theta = matrix.col(direction_map[two_theta_direction])

        R = two_theta.axis_and_angle_as_r3_rotation_matrix(two_theta_angle, deg=True)

        detector = DetectorFactory.make_detector(
            DetectorFactory.sensor(sensor),
            (R * fast),
            (R * slow),
            (R * origin),
            pixel_size,
            image_size,
            trusted_range,
            px_mm,
            mu=mu,
            gain=gain,
            pedestal=pedestal,
            identifier=identifier,
        )

        detector.mask = mask
        return detector

    @staticmethod
    def complex(
        sensor,
        origin,
        fast,
        slow,
        pixel,
        size,
        trusted_range=(0.0, 0.0),
        mask=[],
        px_mm=None,
        mu=0.0,
        gain=None,
        pedestal=None,
        identifier="",
    ):
        """A complex detector model, where you know exactly where everything
        is. This is useful for implementation of the Rigaku Saturn header
        format, as that is exactly what is in there. Origin, fast and slow are
        vectors in the CBF reference frame, pixel is the dimensions as a tuple
        as is size."""

        assert len(origin) == 3
        assert len(fast) == 3
        assert len(slow) == 3
        assert len(pixel) == 2
        assert len(size) == 2

        detector = DetectorFactory.make_detector(
            DetectorFactory.sensor(sensor),
            fast,
            slow,
            origin,
            pixel,
            size,
            trusted_range,
            px_mm,
            mu=mu,
            gain=gain,
            pedestal=pedestal,
            identifier=identifier,
        )
        detector[0].mask = mask
        return detector

    @staticmethod
    def imgCIF(cif_file, sensor):
        """Initialize a detector model from an imgCIF file."""

        cbf_handle = pycbf.cbf_handle_struct()
        cbf_handle.read_file(cif_file.encode(), pycbf.MSG_DIGEST)

        return DetectorFactory.imgCIF_H(cbf_handle, sensor)

    @staticmethod
    def imgCIF_H(cbf_handle, sensor):
        """Initialize a detector model from an imgCIF file handle, where it
        is assumed that the file has already been read."""

        cbf_detector = cbf_handle.construct_detector(0)

        pixel = (
            cbf_detector.get_inferred_pixel_size(1),
            cbf_detector.get_inferred_pixel_size(2),
        )

        origin = tuple(cbf_detector.get_pixel_coordinates(0, 0))
        fast = cbf_detector.get_detector_axis_fast()
        slow = cbf_detector.get_detector_axis_slow()
        size = tuple(reversed(cbf_handle.get_image_size(0)))

        try:
            min_trusted_value = find_underload_value(cbf_handle)
        except Exception:
            try:
                # By convention, if underload is not set, then assume the minimum
                # trusted pixel is 1 more than the undefined pixel
                min_trusted_value = find_undefined_value(cbf_handle) + 1
            except Exception:
                min_trusted_value = 0

        try:
            # In imgCIF overload means the pixel is saturated and hence untrusted
            # https://www.iucr.org/__data/iucr/cifdic_html/2/cif_img.dic/Iarray_intensities.overload.html
            max_trusted_value = (cbf_handle.get_overload(0) - 1) * dxtbx_overload_scale
        except Exception:
            max_trusted_value = 1.0e6

        trusted_range = (min_trusted_value, max_trusted_value)

        gain = find_gain_value(cbf_handle)

        cbf_detector.__swig_destroy__(cbf_detector)
        del cbf_detector

        return DetectorFactory.make_detector(
            DetectorFactory.sensor(sensor),
            fast,
            slow,
            origin,
            pixel,
            size,
            trusted_range,
            gain=gain,
        )

    @staticmethod
    def sensor(name):
        """Return the correct sensor token for a given name, for example:

        ccd, CCD
        image_plate, IMAGE_PLATE
        pad, PAD

        to the appropriate static token which will be used as a handle
        everywhere else in this. Also allow existing token to be passed in."""

        if detector_helper_sensors.check_sensor(name):
            return name

        if name is None:
            return detector_helper_sensors.SENSOR_UNKNOWN

        if name.upper() == "PAD":
            return detector_helper_sensors.SENSOR_PAD
        elif name.upper() == "CCD":
            return detector_helper_sensors.SENSOR_CCD
        elif name.upper() == "IMAGE_PLATE":
            return detector_helper_sensors.SENSOR_IMAGE_PLATE
        elif name.upper() == "UNKNOWN":
            return detector_helper_sensors.SENSOR_UNKNOWN

        raise RuntimeError("name %s not known" % name)



from __future__ import annotations

import os
import shutil
import warnings

import h5py
import numpy as np
import pytest

from dxtbx.command_line.dlsnxs2cbf import parser, run
from dxtbx.format.FormatCBFMiniEigerDLS16MSN160 import FormatCBFMiniEigerDLS16MSN160
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.util.dlsnxs2cbf import make_cbf


@pytest.mark.xfail(reason="Broken for old data while collecting new data")
def test_dlsnxs2cbf(dials_data, tmp_path, capsys):
    screen = dials_data("thaumatin_eiger_screen", pathlib=True)
    master = screen / "Therm_6_1_master.h5"
    run([str(master), "junk_%04d.cbf"])

    output_files = ["junk_%04d.cbf" % j for j in (1, 2, 3)]
    captured = capsys.readouterr()
    for file in output_files:
        assert file.encode("latin-1") in captured.out
        assert tmp_path.joinpath(file).is_file()

    expts = ExperimentListFactory.from_filenames(
        str(tmp_path / file) for file in output_files
    )
    assert all(
        imgset.get_format_class() == FormatCBFMiniEigerDLS16MSN160
        for imgset in expts.imagesets()
    )

    with h5py.File(master) as fh:
        for i, imgset in enumerate(expts.imagesets()):
            original = fh["/entry/data/data_000001"][i][()]
            sel = np.where(original < original.max())
            np.testing.assert_equal(
                fh["/entry/data/data_000001"][i][sel],
                imgset.get_raw_data(0)[0].as_numpy_array()[sel],
            )


@pytest.mark.xfail(reason="The dependency chain is broken")
@pytest.mark.parametrize("remove_axis", ["phi", "chi"], ids=["no phi", "no chi"])
def test_dlsnxs2cbf_deleted_axis(dials_data, tmp_path, remove_axis):
    """Check that a master file without φ or χ axes processes OK."""
    screen = dials_data("thaumatin_eiger_screen", pathlib=True)
    master = "Therm_6_1_master.h5"
    links = (path.name for path in screen.glob("*") if path.name != master)
    try:
        for name in links:
            (tmp_path / name).symlink_to(screen / name)
    except OSError:
        warnings.warn(
            "Copying files where unable to symlink. On Windows, Administrators"
            " or users with Developer Mode can create symlinks freely."
        )
        for name in links:
            shutil.copy(os.fspath(screen / name), os.fspath(tmp_path))
    shutil.copy(screen / master, tmp_path / master)

    with h5py.File(tmp_path / master, "r+") as f:
        del f[f"entry/sample/transformations/{remove_axis}"]

    make_cbf(tmp_path / master, template=str(tmp_path / "image_%04d.cbf"))


def test_dlsnxs2cbf_help(capsys):
    with pytest.raises(SystemExit):
        run(["-h"])
    captured = capsys.readouterr()
    assert parser.description.splitlines()[0] in captured.out
    assert "usage: dxtbx.dlsnxs2cbf" in captured.out



from __future__ import annotations

import sys

from iotbx.detectors.bruker import BrukerImage

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format


class FormatBruker(Format):
    """cctbx has no authoritative sources describing the Bruker format.
    Positive identification:  when listing the header out in 16-byte chunks,
    there are numerous chunks of the form
    KEYNAME:xxxxxxxx
    KEYNAME:xxxxxxxx
    ...in other words a 7-upper case character (or space) keyname followed by
    a colon, starting at the beginning of every fifth chunk.
    Here we will take a series of 12 of these in the first 1024 characters
    as a positive fingerprint of Bruker format.
    """

    @staticmethod
    def read_header_lines(image_path, max_bytes=512 * 50):
        """Attempt to read the whole Bruker header into a list of lines of 80 char
        length. Finish reading lines when either the amount of data determined by
        HDRBLKS is consumed, or if max_bytes is reached, or a line contains a
        non-ASCII character. HDRBLKS is interpreted as being an integer number
        of 512 byte blocks. HDRBLKS must be divisible by 5, so that the total
        number of bytes in the header is divisible by 80."""

        hdr_lines = []
        max_bytes = (max_bytes // 80) * 80
        with FormatBruker.open_file(image_path, "rb") as f:
            while True:

                # read a line and look for HDRBLKS
                line = f.read(80)
                if not line:
                    break
                try:
                    line = line.decode("ascii")
                except UnicodeDecodeError:
                    break
                if line.startswith("HDRBLKS"):
                    try:
                        val = int(line.split(":", 1)[1])
                        if val % 5 == 0:
                            max_bytes = min(val * 512, max_bytes)
                    except ValueError:
                        pass

                # looks like a valid line
                hdr_lines.append(line)

                if max_bytes is not None and f.tell() >= max_bytes:
                    break

        return hdr_lines

    @staticmethod
    def parse_header(header_lines):
        header_dic = {}

        for l in header_lines:
            separator = l.find(":")
            if separator == -1:
                continue
            k = l[:separator].strip()
            v = l[separator + 1 :].strip()
            if k in header_dic:
                header_dic[k] = header_dic[k] + "\n" + v
            else:
                header_dic[k] = v

        return header_dic

    @staticmethod
    def understand(image_file):
        try:
            with FormatBruker.open_file(image_file, "rb") as fh:
                tag = fh.read(1024).decode("latin-1", "replace")
        except OSError:
            return False
        matches = 0
        for x in range(0, 1024, 80):
            word = tag[x : x + 16]
            if len(word) != 16:
                return False
            if word[7] == ":" and word[0:7].isupper():
                matches += 1
                if matches >= 12:
                    return True
        return False

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        super().__init__(str(image_file), **kwargs)

    def detectorbase_start(self):
        pass

    def _start(self):
        """Open the image file, read the image header, copy the key / value
        pairs into an internal dictionary self._header_dictionary along with
        the length of the header in bytes self._header_size."""
        self.detectorbase = BrukerImage(self._image_file)
        # self.detectorbase.readHeader() #unnecessary for the Bruker specialization

    def _goniometer(self):
        if self.detectorbase.parameters["OSC_RANGE"] > 0:
            return self._goniometer_factory.single_axis()
        else:
            return self._goniometer_factory.single_axis_reverse()

    def _detector(self):
        """Return a model for a simple detector"""

        # At present, ignore non-zero two theta for the dxtbx model
        # XXX Return to this issue later.
        # twotheta = self.detectorbase.parameters["TWOTHETA"]

        return self._detector_factory.simple(
            sensor="CCD",
            distance=self.detectorbase.distance,
            beam_centre=(self.detectorbase.beamx, self.detectorbase.beamy),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(self.detectorbase.pixel_size, self.detectorbase.pixel_size),
            image_size=(self.detectorbase.size1, self.detectorbase.size2),
            trusted_range=(0, self.detectorbase.saturation),
            mask=[],
        )  # a list of dead rectangles

    def _beam(self):
        """Return a simple model for the beam."""

        return self._beam_factory.simple(self.detectorbase.wavelength)

    def _scan(self):
        """Return the scan information for this image."""

        return self._scan_factory.single_file(
            filename=self._image_file,
            # It's not at all clear how to recover the exposure time from the header
            # or even whether it is recorded.
            # XXX Here it will simply be set to a default number.
            exposure_times=1,
            osc_start=self.detectorbase.osc_start,
            osc_width=abs(self.detectorbase.deltaphi),
            epoch=None,
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatBruker.understand(arg))



"""
An implementation of the CBF image reader for Pilatus images, from the Pilatus
6M SN 100 currently on Diamond I04.
"""


from __future__ import annotations

import binascii
import sys

import libtbx
from scitbx.array_family import flex

from dxtbx.ext import uncompress
from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus
from dxtbx.masking import GoniometerMaskerFactory


class FormatCBFMiniPilatusDLS6MSN100(FormatCBFMiniPilatus):
    """A class for reading mini CBF format Pilatus images for 6M SN 100 @ DLS."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS" in record
                and "S/N 60-0100 Diamond" in header
            ):
                return True

        return False

    @staticmethod
    def has_dynamic_shadowing(**kwargs):
        dynamic_shadowing = kwargs.get("dynamic_shadowing", False)
        if dynamic_shadowing in (libtbx.Auto, "Auto"):
            return True
        return dynamic_shadowing

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment."""

        self._dynamic_shadowing = self.has_dynamic_shadowing(**kwargs)
        super().__init__(image_file, **kwargs)

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header, though for miniCBF
        there are limited options for this."""

        if "Phi" in self._cif_header_dictionary:
            phi_value = float(self._cif_header_dictionary["Phi"].split()[0])
        else:
            phi_value = 0.0

        if "Kappa" in self._cif_header_dictionary:
            kappa_value = float(self._cif_header_dictionary["Kappa"].split()[0])
        else:
            kappa_value = 0.0

        if "Omega" in self._cif_header_dictionary:
            omega_value = float(self._cif_header_dictionary["Omega"].split()[0])
        else:
            omega_value = 0.0

        phi = (1.0, 0.0, 0.0)
        kappa = (0.914, 0.279, -0.297)
        omega = (1.0, 0.0, 0.0)
        axes = flex.vec3_double((phi, kappa, omega))
        angles = flex.double((phi_value, kappa_value, omega_value))
        names = flex.std_string(("GON_PHI", "GON_KAPPA", "GON_OMEGA"))
        return self._goniometer_factory.make_multi_axis_goniometer(
            axes, angles, names, scan_axis=2
        )

    def get_goniometer_shadow_masker(self, goniometer=None):
        return GoniometerMaskerFactory.mini_kappa(goniometer)

    def _read_cbf_image(self):
        start_tag = binascii.unhexlify("0c1a04d5")

        with self.open_file(self._image_file, "rb") as fh:
            data = fh.read()
        data_offset = data.find(start_tag) + 4
        cbf_header = self._parse_cbf_header(
            data[: data_offset - 4].decode("ascii", "ignore")
        )

        pixel_values = uncompress(
            packed=data[data_offset : data_offset + cbf_header["size"]],
            fast=cbf_header["fast"],
            slow=cbf_header["slow"],
        )

        return pixel_values


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniPilatusDLS6MSN100.understand(arg))



from __future__ import annotations

import sys

import h5py
import numpy as np

from cctbx import factor_ev_angstrom
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.format.Format import Format
from dxtbx.format.FormatHDF5 import FormatHDF5
from dxtbx.format.FormatStill import FormatStill


class FormatHDF5Sacla(FormatHDF5, FormatStill):
    """
    Class for reading SACLA images created by the DataConvert SACLA
    script (this script lives on the SACLA hpc).

    This assumes the argument -reconstr was passed to
    DataConvert in order to Reconstruct the image.

    Also, this processes only a single run's worth of data in the hdf5
    """

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file, "r") as h5_handle:
            understood = False
            if "file_info" in h5_handle and "run_number_list" in h5_handle["file_info"]:
                run_grp = FormatHDF5Sacla._get_run_h5group(h5_handle)
                if "detector_2d_assembled_1" in list(run_grp.keys()):
                    understood = True
            return understood

    def _start(self):
        self._h5_handle = h5py.File(self.get_image_file(), "r")
        self._run = FormatHDF5Sacla._get_run_h5group(self._h5_handle)
        event_info = self._run["event_info"]
        tag_number_list = event_info["tag_number_list"]
        self._images = ["tag_%d" % tag for tag in tag_number_list]

    @staticmethod
    def _get_run_h5group(h5_handle):
        """returns the first run group found"""
        file_info = h5_handle["file_info"]
        run_number_list = file_info["run_number_list"]
        run_str = "run_%d" % run_number_list[0]
        return h5_handle[run_str]

    def _detector(self, index=None):
        # Get the pixel and image size
        detector_2d_assembled_1 = self._run["detector_2d_assembled_1"]
        detector_info = detector_2d_assembled_1["detector_info"]
        pixel_size = (
            detector_info["pixel_size_in_micro_meter"][0] / 1000,
            detector_info["pixel_size_in_micro_meter"][1] / 1000,
        )
        tag = detector_2d_assembled_1[self._images[0]]
        data = tag["detector_data"][()]

        # detector_image_size is fast-, slow- , just in case the dataset is ever non-square
        image_size = (data.shape[1], data.shape[0])
        trusted_range = (0, 200000)

        # Initialise detector frame
        fast = matrix.col((1.0, 0.0, 0.0))
        slow = matrix.col((0.0, -1.0, 0.0))
        orig = matrix.col(
            (
                -image_size[0] * pixel_size[0] / 2,
                image_size[1] * pixel_size[1] / 2,
                -100.0,
            )
        )
        # Make the detector
        return self._detector_factory.make_detector(
            "", fast, slow, orig, pixel_size, image_size, trusted_range
        )

    def _beam(self, index=None):
        run_info = self._run["run_info"]
        sacla_config = run_info["sacla_config"]
        eV = sacla_config["photon_energy_in_eV"].value

        return self._beam_factory.simple(factor_ev_angstrom / eV)

    def get_num_images(self):
        return len(self._images)

    def get_raw_data(self, index=0):
        detector_2d_assembled_1 = self._run["detector_2d_assembled_1"]
        tag = detector_2d_assembled_1[self._images[index]]
        return flex.double(tag["detector_data"].value.astype(np.float64))

    def get_detectorbase(self, index=None):
        raise NotImplementedError

    def get_image_file(self, index=None):
        return Format.get_image_file(self)

    def get_detector(self, index=None):
        return self._detector_instance

    def get_beam(self, index=None):
        return self._beam_instance


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatHDF5Sacla.understand(arg))



"""
An implementation of the SMV image reader for ADSC images. Inherits from
FormatSMVADSC, customised for beamline 8.3.1 at the ALS where J. Holton uses
two-theta offsets in the vertical direction, as well as idiosyncratic ways
of recording the beam centre... which work fine for ADXV...
"""

from __future__ import annotations

import math

from dxtbx.format.FormatSMVADSCSN import FormatSMVADSCSN


class FormatSMVADSCSN926(FormatSMVADSCSN):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this, for instrument number 926."""

    @staticmethod
    def understand(image_file):
        """Check to see if this is ADSC SN 926."""

        # check this is detector serial number 926 (or 907)

        size, header = FormatSMVADSCSN.get_smv_header(image_file)

        if int(header["DETECTOR_SN"]) not in [926, 907]:
            return False

        return True

    def _detector(self):
        """Return a model for a simple detector, allowing for the installation
        on on a two-theta stage. Assert that the beam centre is provided in
        the Mosflm coordinate frame."""

        distance = float(self._header_dictionary["DISTANCE"])
        if "DENZO_X_BEAM" in self._header_dictionary:
            beam_x = float(self._header_dictionary["DENZO_X_BEAM"])
            beam_y = float(self._header_dictionary["DENZO_Y_BEAM"])
        else:
            beam_x = float(self._header_dictionary["BEAM_CENTER_X"])
            beam_y = float(self._header_dictionary["BEAM_CENTER_Y"])
        pixel_size = float(self._header_dictionary["PIXEL_SIZE"])
        image_size = (
            float(self._header_dictionary["SIZE1"]),
            float(self._header_dictionary["SIZE2"]),
        )
        if "TWOTHETA" in self._header_dictionary:
            two_theta = float(self._header_dictionary["TWOTHETA"])
        else:
            two_theta = 0

        # now correct for some idiosyncrasies...
        # two-theta included in beam centre - so remove this
        beam_y += distance * math.tan(two_theta * math.pi / 180.0)

        return self._detector_factory.two_theta(
            "CCD",
            distance,
            (beam_y, beam_x),
            "+x",
            "-y",
            "+x",
            two_theta,
            (pixel_size, pixel_size),
            image_size,
            self._adsc_trusted_range(),
            [],
            gain=self._adsc_module_gain(),
        )



# LIBTBX_SET_DISPATCHER_NAME dev.dxtbx.debug_memory

from __future__ import annotations

import argparse
import resource

import dxtbx
import dxtbx.util


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser(
        description="Test memory usage by repeatedly loading an image"
    )
    parser.add_argument("image_frame", help="An image to reload repeatedly")
    options = parser.parse_args(args)

    powers = [2**n for n in range(20)]

    for j in range(powers[-1] + 1):
        dxtbx.load(options.image_frame)
        mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        if j in powers:
            print(j, mem)


if __name__ == "__main__":
    run()
